/*================================================================ 
 	Copyright 2008, Yoav Goldberg.
 	
 	This file is part of splitSVM.
 
    splitSVM is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    splitSVM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with splitSVM.  If not, see http://www.gnu.org/licenses/.
     
 ================================================================*/
package yg.classifier.splitsvm;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import gnu.trove.*;
import java.util.Arrays;

import yg.classifier.fmap.FeatureMap;
import yg.classifier.splitsvm.io.FastMulticlassModelReader;
import yg.classifier.splitsvm.io.FastMulticlassZippedModelReader;
import yg.classifier.splitsvm.io.KernelData;

import libsvm.svm_node;

public class FastMulticlassModel {
	private static boolean alreadyLoaded = false;
	
	public static final int OVR = 1;
	public static final int PAIRWISE = 2;
	
	protected FeatsToVecs featsToVecs;//    = null;
	protected PairWeights pairWeights;//    = null;
	protected SV[]			svs		   = null;
	protected FeatureMap  	fmap       = null;
	
	protected LabelsData labelsData = null;
	
	protected MulticlassResolver multiclassResolver = null;
	
	protected double gamma2 = 2;
	protected double gammasqr = 1;
	protected double coef0;
	
	protected double[] thresholds = null;	
	protected boolean[] isCommonArray = null;
	
	protected double[] scoresToFill = null;
	
	// for internal use of "classify"
	private int[]    seensvsi = null; 
	private int seencurrent = 0;
	
	private int[] _intsLabelsArray = null;
	
	protected int[] rares = null;
	protected int[] commons = null;
	protected int rarec;
	protected int commonc; 
	
	protected String modelFilePrefix;
	
	
	public FastMulticlassModel() {
		if (!FastMulticlassModel.alreadyLoaded) {
			this.splash();
		}
		FastMulticlassModel.alreadyLoaded = true;
	}
	
	private void splash() {
		System.out.println("=====================================================================================");
		System.out.println("                 splitsvm: Fast SVM decoder for NLP applications.");
		System.out.println("-------------------------------------------------------------------------------------");
		System.out.println("                         Ben Gurion University, NLP Group. ");
		System.out.println("                         http://www.cs.bgu.ac.il/~nlpproj  ");
		System.out.println("-------------------------------------------------------------------------------------");
		System.out.println("If you use this software in academic research, please cite:");
		System.out.println("  Yoav Goldberg and Michael Elhadad (2008) splitSVM: Fast, Space-Efficient,"); 
		System.out.println("  non-Heuristic, Polynomial Kernel Computation for NLP Applications. In Proceedings");
		System.out.println("  of the 46th Annual Meeting of the Association for Computational Linguistics. ");
		System.out.println("=====================================================================================");		
	}
	
	public void read(String file_prefix) throws UnsupportedEncodingException, IOException {
		this.read(file_prefix, true);
	}
	
	public void read(String file_prefix, boolean loadFmap) throws UnsupportedEncodingException, IOException {
		FastMulticlassModelReader reader;
		if (file_prefix.endsWith(".spltmod")) {
			reader = new FastMulticlassZippedModelReader(file_prefix);
		} else {
			reader = new FastMulticlassModelReader();	
		}
		
		this.modelFilePrefix = file_prefix;
		// common features -- fill isCommonArray[feature_id]
		TIntHashSet commonFeatures = reader.read_common_features(file_prefix);
		int[] tmparray = commonFeatures.toArray();
		if (tmparray.length == 0) { isCommonArray = new boolean[] {}; } else {
			Arrays.sort(tmparray);
			isCommonArray = new boolean[tmparray[tmparray.length - 1] + 1];
			tmparray = null;
			for (int fid : commonFeatures.toArray()) {
				isCommonArray[fid]=true;
			}		
		} 
		commonFeatures = null;
		
		featsToVecs = reader.read_rare_to_vecs(file_prefix);
		
		pairWeights = reader.read_common_pair_weights(file_prefix);
		
		svs = reader.read_vectors(file_prefix);		
		for (SV sv : svs) { sv.initializeCommon(isCommonArray); }		
		
		rares = new int[svs[0].features().length];
		commons = new int[svs[0].features().length];
		
		if (loadFmap) { fmap = reader.read_features_map(file_prefix); }
		
		thresholds  = reader.read_thresholds(file_prefix);
		scoresToFill = new double[thresholds.length];
	
		multiclassResolver = reader.readMulticlassData(file_prefix);
		labelsData = multiclassResolver.getLabelsData();
		
		//int nr_classifiers = thresholds.length;
		//seensvs = new boolean[svs.length];
		seensvsi = new int[svs.length];
		
		//scores  = new double[nr_classifiers];
		
		KernelData kernel = reader.read_kernel(file_prefix);
		if (kernel.degree != 2) {throw new IOException("unsupported kernel degree " + kernel.degree); }
		gamma2 = 2*kernel.gamma;
		gammasqr = kernel.gamma * kernel.gamma;
		coef0 = kernel.coef0;
		
	}
		
	private int countIntersection(int[] x, int[] y, int leny) {
		int cnt = 0;
		int lenx = x.length;
		int i=0; int j=0;
		while (i < lenx && j < leny) {
			if (x[i]==y[j]) { cnt++; i++; j++;}
			else {
				if (x[i] < y[j]) {i++; }
				else {j++;}
			}
		}	
		return cnt;
	}
	
	/**
	 * Fills the commons[] and rares[] with indexes of common and rare features.
	 * Sets the commonc and rarec to hold the number of common and rare features respectively.
	 * 
	 * @param features
	 */
	protected void sepRareAndCommon(svm_node[] features) {
		commonc = 0;
		rarec = 0;
		for (int i=0;i<features.length;i++) {
			int f = features[i].index;
			if (f < isCommonArray.length && isCommonArray[f]) {
				commons[commonc++] = f; 
			} else {				
				rares[rarec++] = f;
			}
		}	
	}
	
	/**
	 * Fills the commons[] and rares[] with indexes of common and rare features.
	 * Sets the commonc and rarec to hold the number of common and rare features respectively.
	 * 
	 * @param features
	 */
	protected void sepRareAndCommon(int[] features) {
		commonc = 0;
		rarec = 0;
		for (int i=0;i<features.length;i++) {
			int f = features[i];
			if (f < isCommonArray.length && isCommonArray[f]) {
				commons[commonc++] = f; 
			} else {				
				rares[rarec++] = f;
			}
		}
	}
		
	
	protected void fillClassifierScores(double[] scores) {
		System.arraycopy(thresholds, 0, scores, 0, thresholds.length);
		
		if (++seencurrent <= 0) { Arrays.fill(seensvsi, 0); seencurrent = 1;}
		for (int k=0;k<rarec;k++) {
			int feat = rares[k];
			for (int svidx : featsToVecs.vecsForFeat(feat)) {
				if (seensvsi[svidx] < seencurrent) {
					seensvsi[svidx] = seencurrent;
					SV sv = svs[svidx];
					int rdot = this.countIntersection(sv._rareFeatures, rares, rarec);					
					int cdot = this.countIntersection(sv._commonFeatures, commons, commonc);
					//score += sv.ai()*((3*rdot*rdot) + (2*rdot*cdot) + (2*rdot));
					//int tmp = ((1*rdot*rdot) + (2*rdot*cdot) + (2*rdot) );
					double tmp = ((gammasqr*rdot*rdot)+(2*gammasqr*rdot*cdot) + (gamma2*rdot*coef0));

					for (TypeClassifierAi cai : sv.classifiersAIs()) {						
						scores[cai.classifierNum] += (tmp * cai.ai);
					}
				} 
			}
		}
		// common pairs  
		for (int i=0;i<commonc;i++) {
			for (int j=i;j<commonc;j++) {
				for (TypeClassifierWeight cw : pairWeights.getWeights(commons[i],commons[j])) {
					scores[cw.classifierId] += cw.w;
				}
			}
		}				
	}
	
	
	public void getClassifiersScores(svm_node[] features, double[] scores) {
		this.sepRareAndCommon(features);
		this.fillClassifierScores(scores);
	}
	
	public void getClassifiersScores(int[] features, double[] scores) {
		this.sepRareAndCommon(features);
		this.fillClassifierScores(scores);
	}
	
	public boolean binaryClassify(int[] features) {
		double[] scores = {0.0};
		this.sepRareAndCommon(features);
		this.fillClassifierScores(scores);
		return scores[0] > 0;
	}
	

	public int classifyToInt(int[] features) {
		//Arrays.fill(scoresToFill, 0);
		this.getClassifiersScores(features, scoresToFill);
			
		return multiclassResolver.resolveToClassifierId(scoresToFill);		
	}
	
	public String classifyToLabels(int[] features) {
		this.getClassifiersScores(features, scoresToFill);
		
		return multiclassResolver.resolveToLabel(scoresToFill);
	}
	
	public String classifyToLabels(svm_node[] features) {
		this.getClassifiersScores(features, scoresToFill);
		
		return multiclassResolver.resolveToLabel(scoresToFill);
	}
	
	public FeatureMap getFeatMap() {
		return this.fmap;		
	}
	public int getNumOfLabels() {
		return labelsData.getLabels().length;		
	}
	
	public String[] getLabels() {
		return labelsData.getLabels();		
	}
	
	public String intToLabel(int labelNum) {
		return this.getLabels()[labelNum];
	}
	
	public String trueLabelForClassifier(int cid) {
		return labelsData.getTrue(cid);
	}
	public int[] getLabelsAsInts() {
		/***** Start MaltParser hack **/
		if (null == _intsLabelsArray) {
			String[] lbls = this.getLabels();
			_intsLabelsArray = new int[lbls.length];
			for (int i = 0;i<lbls.length;i++) {
				_intsLabelsArray[i] = Integer.parseInt(lbls[i]);
			}			
		} 
		return _intsLabelsArray;
		/***** End MaltParser hack **/
	}
	 
	

	

}
