/*================================================================ 
 	Copyright 2008, Yoav Goldberg.
 	
 	This file is part of splitSVM.
 
    splitSVM is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    splitSVM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with splitSVM.  If not, see http://www.gnu.org/licenses/.
     
 ================================================================*/
package yg.classifier.splitsvm.io;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.Vector;

import yg.classifier.fmap.FeatureMap;
import yg.classifier.splitsvm.FastMulticlassModel;
import yg.classifier.splitsvm.FeatsToVecs;
import yg.classifier.splitsvm.LabelsData;
import yg.classifier.splitsvm.OVRMulticlassResolver;
import yg.classifier.splitsvm.PairWeights;
import yg.classifier.splitsvm.PairwiseMulticlassResolver;
import yg.classifier.splitsvm.SV;
import yg.classifier.splitsvm.MulticlassResolver;

import gnu.trove.TIntHashSet;

public class FastMulticlassModelReader {

	public static final String EXT_AIS = ".ais";
	public static final String EXT_KERNEL = ".kernel";
	public static final String MODELMAIN_EXT = ".classifiers";
	public static final String EXT_COMMONW_BINARY = ".bcommonpairs";
	public static final String EXT_VECS_BINARY = ".bvecs";
	public static final String EXT_AIS_BINARY = ".bais";

	protected InputStream getInputStream(String prefix, String extension) throws FileNotFoundException, IOException {
		return new FileInputStream(prefix + extension);
	}
	
	/*
	public static LabelsData readLabelsData(String filename_prefix) throws IOException {
		File f = new File(filename_prefix + MODELMAIN_EXT);
		BufferedReader b = new BufferedReader(new InputStreamReader(new FileInputStream(f)) );
		String[] type = b.readLine().split(":"); // first line -- type
		int multiclass = 0;
		//String otherLabel = null;
		if (type[1].equals("ovr")) {
			multiclass = FastMulticlassModel.OVR;
			//otherLabel = type[2]; 
		} else if (type[1].equals("pairwise")) {
			multiclass = FastMulticlassModel.PAIRWISE;
		} else {
			throw new IOException("unknown model type. should be type:ovr or type:pairwise");
		}
		
		//String[] labels = b.readLine().split("\\s+");
		String[] labels = b.readLine().split(" ");
		String line = null;
		Vector<String> lines = new Vector<String>();
		while (null != (line = b.readLine())) {
			lines.add(line);
		}
		String[] trueLabels = new String[lines.size()];
		String[] falseLabels = new String[lines.size()];
		for (String ln : lines) {
			//String[] split = line.split("\\s+");
			String[] split = ln.split(" ");
			int cid = Integer.parseInt(split[0]);
			trueLabels[cid]   = split[1];
			falseLabels[cid]  = split[2];								
		}		
		return new LabelsData(labels, trueLabels, falseLabels);
	}*/
	
	public MulticlassResolver readMulticlassData(String filename_prefix) throws IOException {
		InputStream is = getInputStream(filename_prefix, MODELMAIN_EXT);
		BufferedReader b = new BufferedReader(new InputStreamReader(is) );
		
		String[] type = b.readLine().split(":"); // first line -- type
		int multiclass = 0;
		String otherLabel = null; // label for the "other" class in "one vs others" model
		if (type[1].equals("ovr")) {
			multiclass = FastMulticlassModel.OVR;
			otherLabel = type[2]; 
		} else if (type[1].equals("pairwise")) {
			multiclass = FastMulticlassModel.PAIRWISE;
		} else {
			throw new IOException("unknown model type. should be type:ovr or type:pairwise");
		}
		
		//String[] labels = b.readLine().split("\\s+");
		String[] labels = b.readLine().split(" ");
		String line = null;
		Vector<String> lines = new Vector<String>();
		while (null != (line = b.readLine())) {
			lines.add(line);
		}
		String[] trueLabels = new String[lines.size()];
		String[] falseLabels = new String[lines.size()];
		for (String ln : lines) {
			//String[] split = line.split("\\s+");
			String[] split = ln.split(" ");
			int cid = Integer.parseInt(split[0]);
			trueLabels[cid]   = split[1];
			falseLabels[cid]  = split[2];								
		}
		LabelsData ld = new LabelsData(labels, trueLabels, falseLabels);
		if (multiclass == FastMulticlassModel.OVR) {			
			return new OVRMulticlassResolver(ld, otherLabel);
		} else {
			return new PairwiseMulticlassResolver(ld);
		}		
	}
	
	/**
	 * returns a hashset of common-feature ids (Over ALL svs, in all the classifiers) 
	 * @param file_prefix prefix of filename to load
	 * @return set of common-feature ids
	 */
	public TIntHashSet read_common_features(String file_prefix) throws UnsupportedEncodingException, IOException {
		return ModelReader.read_common_features(getInputStream(file_prefix, ModelReader.EXT_COMMON));
	}

	/**
	 * a reverse index map from feature-id to SV-id, over ALL features and all vectors in all the classifiers.
	 * @param file_prefix prefix of filename to load
	 * @return
	 */
	public FeatsToVecs read_rare_to_vecs(String file_prefix) throws UnsupportedEncodingException, IOException {
		return ModelReader.read_rare_to_vecs(getInputStream(file_prefix, ModelReader.EXT_RARERIDX));
		
	}

	/**
	 * PairWeights object is initialized with add(int f1, int f2, int classifierNum, double w) 
	 * over all pairs in all classifiers (and non-0 w values)
	 * @param file_prefix prefix of filename to load
	 * @return
	 * @throws IOException 
	 * @throws NumberFormatException 
	 */
	protected PairWeights read_common_pair_weights_binary(String file_prefix) throws IOException {
		InputStream is = getInputStream(file_prefix, EXT_COMMONW_BINARY); 
		DataInputStream in = new DataInputStream(new BufferedInputStream(is));
		
		PairWeights pw = new PairWeights();
		
		while (true) {
			try {
				int f1 = in.readInt();
				int f2 = in.readInt();
				short num = in.readShort();
				for (int i=0;i<num;i++) {
					short id = in.readShort();
					double val = in.readDouble();
					pw.add(f1, f2, id, val);
				}				
			}
			catch (EOFException e) {
				break;
			}
		}
		pw.doneAddingWeights();		
		return pw;
	}
	
	/**
	 * PairWeights object is initialized with add(int f1, int f2, int classifierNum, double w) 
	 * over all pairs in all classifiers (and non-0 w values)
	 * @param file_prefix prefix of filename to load
	 * @return
	 * @throws IOException 
	 * @throws NumberFormatException 
	 */
	public PairWeights read_common_pair_weights(String file_prefix) throws IOException {
		InputStream is;
		try {
			is = getInputStream(file_prefix, ModelReader.EXT_COMMONW);
		} catch (FileNotFoundException e) {
			return read_common_pair_weights_binary(file_prefix);
		} 
		
		BufferedReader br = new BufferedReader(new InputStreamReader(is) );
		
		String line = null;
		PairWeights pw = new PairWeights();
		
		while ((line = br.readLine()) != null) {
			//String[] parts = line.split("\\s+");
			//String[] parts = line.split(" ");
			//int f1 = Integer.parseInt(parts[0]);
			//int f2 = Integer.parseInt(parts[1]);
			//for (int i = 2; i<parts.length;i++) {				
			//	String[] cid_w = parts[i].split(":");
			//	pw.add(f1,f2, Integer.parseInt(cid_w[0]), Double.parseDouble(cid_w[1]));
			StringTokenizer tok = new StringTokenizer(line," :");		
			int f1 = Integer.parseInt(tok.nextToken());
			int f2 = Integer.parseInt(tok.nextToken());
			while (tok.hasMoreTokens()) {
				pw.add(f1,f2,Integer.parseInt(tok.nextToken()), Double.parseDouble(tok.nextToken()));
			}
					
		}
		pw.doneAddingWeights();
		br.close();
		return pw;
	}
	
	/**
	 * read a list of all SVs in all classifiers.  
	 * !!! Array index is consistant with indexes retyrned by read_rare_to_vecs
	 * SV will include a separate ai() value for each classifier with non-0 ai
	 * 
	 * @param file_prefix prefix of filename to load
	 * @throws IOException 
	 */
	public SV[] read_vectors(String file_prefix) throws IOException {
		InputStream isv;
		BufferedReader brv;
		
		InputStream isa;
		BufferedReader bra;
		
		try {
			isv = getInputStream(file_prefix, ModelReader.EXT_VECS);
			brv = new BufferedReader(new InputStreamReader(isv) );
		
			isa = getInputStream(file_prefix, EXT_AIS);
			bra = new BufferedReader(new InputStreamReader(isa) );
		} catch (FileNotFoundException e) {
			return read_vectors_binary(file_prefix);
		}
		
		String linev = null;
		String linea = null;
		ArrayList<SV> svs = new ArrayList<SV>();		
		while ((linev = brv.readLine()) != null) {
			linea = bra.readLine();
			//int[] features = ModelReader.strs_to_ints(linev.split("\\s+"));
			int[] features = ModelReader.strs_to_ints(linev.split(" "));
			//String[] cids_ais = linea.split("\\s+");
			String[] cids_ais = linea.split(" ");
			int[] cids = new int[cids_ais.length];
			double[] ais = new double[cids_ais.length];
			for (int k=0;k<cids_ais.length;k++) {
				String[] cid_ai = cids_ais[k].split(":");
				cids[k] = Integer.parseInt(cid_ai[0]);
				ais[k] = Double.parseDouble(cid_ai[1]);
			}
			svs.add(new SV(ais, cids, features));
		}
		brv.close();
		bra.close();
		return svs.toArray(new SV[0]);
	}

	/**
	 * read a list of all SVs in all classifiers.  
	 * !!! Array index is consistant with indexes retyrned by read_rare_to_vecs
	 * SV will include a separate ai() value for each classifier with non-0 ai
	 * 
	 * @param file_prefix prefix of filename to load
	 * @throws IOException 
	 */
	protected SV[] read_vectors_binary(String file_prefix) throws IOException {		
		InputStream isv = getInputStream(file_prefix, EXT_VECS_BINARY);
		DataInputStream inv = new DataInputStream(new BufferedInputStream(isv) );
		
		InputStream isa = getInputStream(file_prefix, EXT_AIS_BINARY);
		DataInputStream ina = new DataInputStream(new BufferedInputStream(isa) );
		
		ArrayList<SV> svs = new ArrayList<SV>();
		int c = 0;
		while (true) {
			try {
				c++;
				short num_feats = inv.readShort();
				int[] features = new int[num_feats]; 
				for (int i=0; i<num_feats;i++) {
					features[i] = inv.readInt();
				}				
				
				short num_classifiers = ina.readShort();
				int[] cids = new int[num_classifiers];
				double[] ais = new double[num_classifiers];
				for (int i=0;i<num_classifiers;i++) {
					cids[i] = ina.readShort();
					ais[i]  = ina.readDouble();
				}
				
				svs.add(new SV(ais, cids, features));				
			} catch (EOFException e) {				
				break;
			}
		}
		ina.close();
		inv.close();
		return svs.toArray(new SV[0]);
	}
	
	/**
	 * returns the -threshold (that is, 0 minus the threshold) for each classifier, in order.
	 * @param file_prefix prefix of filename to load
	 * @return
	 * @throws IOException 
	 */
	public double[] read_thresholds(String file_prefix) throws IOException {
		InputStream is = getInputStream(file_prefix, ModelReader.EXT_THRS);
		BufferedReader br = new BufferedReader(new InputStreamReader(is) );
		
		//String[] parts = br.readLine().split("\\s+");
		String[] parts = br.readLine().split(" ");
		double[] thrs = new double[parts.length];
		for (int i=0;i<thrs.length;i++) {
			thrs[i] = -Double.parseDouble(parts[i]);
		}
		return thrs;
	}


	public KernelData read_kernel(String file_prefix)  throws IOException {		
		InputStream is = getInputStream(file_prefix, EXT_KERNEL);
		BufferedReader br = new BufferedReader(new InputStreamReader(is) );
		KernelData k = new KernelData(
				//Integer.parseInt(br.readLine()), // degree
				(int)Double.parseDouble(br.readLine()), // degree
				Double.parseDouble(br.readLine()), // gamma
				Double.parseDouble(br.readLine())); // coef0
		br.close();
		return k;
	}

	public FeatureMap read_features_map(String file_prefix) throws IOException {
		return ModelReader.read_features_map(getInputStream(file_prefix, ModelReader.EXT_FMAP));		
	}

}
