/*================================================================ 
 	Copyright 2008, Yoav Goldberg.
 	
 	This file is part of splitSVM.
 
    splitSVM is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    splitSVM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with splitSVM.  If not, see http://www.gnu.org/licenses/.
     
 ================================================================*/
package yg.classifier.splitsvm.io;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;

import yg.classifier.fmap.FeatureMap;
import yg.classifier.splitsvm.FeatsToVecs;
import yg.classifier.splitsvm.PairWeights;
import yg.classifier.splitsvm.SV;

import gnu.trove.TIntHashSet;

public class ModelReader {
	public static final String EXT_COMMON = ".common";
	public static final String EXT_RARERIDX = ".rarerindex";
	public static final String EXT_COMMONW = ".commonpairs";
	public static final String EXT_VECS    = ".vecs";
	public static final String EXT_FMAP    = ".fmap";
	public static final String EXT_THRS    = ".thrs";
	
	protected static int[] strs_to_ints(String[] strs) {
		int[] res = new int[strs.length];
		for (int i=0;i<strs.length;i++) {
			res[i] = Integer.parseInt(strs[i]);
		}
		return res;		
	}
	
	
	public static TIntHashSet read_common_features(String file_prefix) throws UnsupportedEncodingException, IOException {
		return read_common_features(new FileInputStream(file_prefix + EXT_COMMON));
	}
	
	public static TIntHashSet read_common_features(InputStream is) throws UnsupportedEncodingException, IOException {		
		BufferedReader br = new BufferedReader(new InputStreamReader(is,"cp1255"));
		String line = br.readLine();
		int[] commons;
		if (line != null)
		    //commons = strs_to_ints(line.split("\\s+"));
			commons = ModelReader.strs_to_ints(line.split(" "));
		else
			commons = new int[0];
		br.close();
		return new TIntHashSet(commons);		
	}
	
	public static FeatsToVecs read_rare_to_vecs(String file_prefix) throws UnsupportedEncodingException, IOException {
		String filename = file_prefix + EXT_RARERIDX;
		return read_rare_to_vecs(new FileInputStream(filename));
	}
	
	public static FeatsToVecs read_rare_to_vecs(InputStream is) throws UnsupportedEncodingException, IOException {		
		BufferedReader br = new BufferedReader(new InputStreamReader(is,"cp1255"));
		String line = null;
		int max_feature_index = Integer.parseInt(br.readLine());
		TIntHashSet[] featToVecs = new TIntHashSet[max_feature_index+1]; 
		while ((line = br.readLine()) != null) {
			//String[] parts = line.split("\\s+");
			String[] parts = line.split(" ");
			int fnum = Integer.parseInt(parts[0]);
			int[] vecsnums = ModelReader.strs_to_ints(parts[1].split(":"));
			featToVecs[fnum] = new TIntHashSet(vecsnums);
		}
		br.close();
		return new FeatsToVecs(featToVecs); 		
	}
	
	public static PairWeights read_common_pair_weights(String file_prefix) throws UnsupportedEncodingException, IOException {
		String filename = file_prefix + EXT_COMMONW;
		BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(filename),"cp1255"));
		String line = null;
		PairWeights pw = new PairWeights();
		while ((line = br.readLine()) != null) {
			//String[] parts = line.split("\\s+");
			String[] parts = line.split(" ");
			int f1 = Integer.parseInt(parts[0]);
			int f2 = Integer.parseInt(parts[1]);
			double w = Double.parseDouble(parts[2]);
			pw.add(f1,f2,w);
		}
		br.close();
		return pw;
	}
	public static SV[] read_vectors(String file_prefix) throws UnsupportedEncodingException, IOException {
		String filename = file_prefix + EXT_VECS;
		BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(filename),"cp1255"));
		String line = null;
		ArrayList<SV> svs = new ArrayList<SV>();		
		while ((line = br.readLine()) != null) {			
			//String[] parts = line.split("\\s+",2);
			String[] parts = line.split(" ",2);
			double ai = Double.parseDouble(parts[0]);
			//int[] features = strs_to_ints(parts[1].split("\\s+"));
			int[] features = strs_to_ints(parts[1].split(" "));
			svs.add(new SV(ai, features));
		}
		br.close();
		return svs.toArray(new SV[0]);
	}
	
	public static FeatureMap read_features_map(String file_prefix) throws UnsupportedEncodingException, IOException {
		String filename = file_prefix + EXT_FMAP;
		return read_features_map(new FileInputStream(filename));
	}

	public static FeatureMap read_features_map(InputStream is) throws UnsupportedEncodingException, IOException {		
		BufferedReader br = new BufferedReader(new InputStreamReader(is,"cp1255"));
		FeatureMap fm = new FeatureMap(br, new FeatureMapReader());
		br.close();
		return fm;
	}

	public static double read_threshold(String file_prefix) throws UnsupportedEncodingException, IOException {
		String filename = file_prefix + EXT_THRS;
		BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(filename),"cp1255"));
		double threshold = Double.parseDouble(br.readLine());
		br.close();
		return threshold;
	}
}
