package yg.tagger;

import gnu.trove.TIntHashSet;
import gnu.trove.TIntIntHashMap;
import gnu.trove.TIntObjectHashMap;

import java.util.IdentityHashMap;

/**
 * 
 * @author yoavg
 * A map from features to IDs.  As the map is sequential, each feature is composed of:
 * symbol  (a string)
 * type    (a number representing the type of the feature, eg, POS, LEX)
 * location (a number representing the location of the feature in relation to the current token (e.g. -1, +2)
 * 
 */
public class SequentialFeatureMap {

	private static final int DYN_FEATURE_TYPE = -1;
	
	private IdentityHashMap<String, TIntObjectHashMap<TIntIntHashMap>> symbolTypeLoc2fid;
	
	private TIntObjectHashMap<TIntHashSet> possibleWindowLocationsForFeatureType;
	
	public SequentialFeatureMap() {
		symbolTypeLoc2fid = new IdentityHashMap<String, TIntObjectHashMap<TIntIntHashMap>>();
		possibleWindowLocationsForFeatureType = new TIntObjectHashMap<TIntHashSet>();
	}
	
	/**
	 * 
	 * @param fmap: an fmap where the keys are in Yamcha format (F:TYPE:LOC:SYM  or T:TYPE:LOC:SYM)
	 */
	static SequentialFeatureMap fromYamchaFeatureMap(yg.classifier.fmap.FeatureMap fmap) {
		SequentialFeatureMap f = new SequentialFeatureMap();
		for (String feature: fmap.features()) {
			int type;
			int loc;
			String symbol;
		
			String[] parts = feature.split(":",4);
			if (parts[0].equals("F")) {
				loc   = Integer.parseInt(parts[1].replace("+",""));
				type  = Integer.parseInt(parts[2]);
				symbol = parts[3];
			} else {
				loc  = Integer.parseInt(parts[1]);
				type = DYN_FEATURE_TYPE;
				symbol = parts[2];
			}
			
			int fid = fmap.getNum(feature); 
			
			f.addFeature(symbol, type, loc, fid);
		}
		return f;
	}

	public void addFeature(String symbol, int type, int loc, int featureId) {
		symbol = symbol.intern();
		if (!symbolTypeLoc2fid.containsKey(symbol)) {
			symbolTypeLoc2fid.put(symbol, new TIntObjectHashMap<TIntIntHashMap>());
		}
		TIntObjectHashMap<TIntIntHashMap> typeLoc2fid = symbolTypeLoc2fid.get(symbol);
		if (!typeLoc2fid.containsKey(type)) {
			typeLoc2fid.put(type, new TIntIntHashMap());
		}
		typeLoc2fid.get(type).put(loc, featureId);
		
		// keep type->locations mapping
		if (!possibleWindowLocationsForFeatureType.containsKey(type)) {
			possibleWindowLocationsForFeatureType.put(type, new TIntHashSet());
		}
		possibleWindowLocationsForFeatureType.get(type).add(loc);
		
		
	}
	
	public int getFeatureId(String symbol, int type, int loc) {
		symbol = symbol.intern();
		if (symbolTypeLoc2fid.containsKey(symbol)) {
			TIntObjectHashMap<TIntIntHashMap> typeLoc2fid = symbolTypeLoc2fid.get(symbol);
			if (typeLoc2fid.containsKey(type)) {
				TIntIntHashMap loc2fid = typeLoc2fid.get(type);
				if (loc2fid.containsKey(loc)) {
					return loc2fid.get(loc);
				}
			}
		}
		//System.out.println("not found:" + loc + ":" + type + ":" + symbol);
		return -1;
	}
	
	public int[] getPossibleWindowLocationsForFeatureType(int type) {
		if (possibleWindowLocationsForFeatureType.containsKey(type)) {
			return possibleWindowLocationsForFeatureType.get(type).toArray();
		} else {
			return new int[0];
		}
	}

	public int dynFeatureType() {
		return DYN_FEATURE_TYPE;		
	}
	
	public int[] getPossibleTypes() {
		return possibleWindowLocationsForFeatureType.keys();
	}
}
