package yg.tagger;

import gnu.trove.TIntHashSet;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import yg.classifier.splitsvm.FastMulticlassModel;

public class SequentialTagger {
	
	final static int DIR_FORWARD = 0;
	final static int DIR_BACKWARD = 1;
	
	FastMulticlassModel model;
	SequentialFeatureMap fmap;	
			
	private ArrayList<FeatureSequence> sequences;
	
	private PositionSeqMaker positionSeqMaker;
	
	private TIntHashSet possibleFeatureTypes;
	
	public SequentialTagger(FastMulticlassModel model) {
		int direction = DIR_FORWARD;
		this.model = model;
		this.fmap = SequentialFeatureMap.fromYamchaFeatureMap(model.getFeatMap());
	
		// TODO: parameter
		if (direction == DIR_FORWARD) {
			this.positionSeqMaker = new Forward(); 
		} else {
			this.positionSeqMaker = new Backward();
		}
		
		possibleFeatureTypes = new TIntHashSet(this.fmap.getPossibleTypes());
	}
	
	/////////////////////////////////////////////////////////////
	
	
	
	public void begin() {		
		sequences = new ArrayList<FeatureSequence>();
		
	}
	
	public void addFeaturesSequence(int TYPE, String[] symbols) {
		if (possibleFeatureTypes.contains(TYPE)) {
			for (int i=0;i<symbols.length;i++) { symbols[i] = symbols[i].intern(); }
			sequences.add(new FeatureSequence(TYPE, symbols));
		}
	}
	
	public void addFeaturesSequence(int TYPE, List<String> symbols) {
		this.addFeaturesSequence(TYPE, symbols.toArray(new String[] {""}));		
	}
	
	/* TODO: this is the usage
	begin();
	addFeaturesSequence(POS, poses);
	addFeaturesSequence(WORD, words);
	addFeaturesSequence(LEMMA, lemmas);
	tag();
	*/
	public ArrayList<String> tag() {		
		int length = sequences.get(0).symbols.length;
	
		int[] positions = positionSeqMaker.range(length - 1);  
			
		ArrayList<String> res = new ArrayList<String>();
		
		TIntHashSet[] featuresPerLocation = new TIntHashSet[length];
		for (int i=0; i<featuresPerLocation.length;i++) {
			featuresPerLocation[i] = new TIntHashSet();
		}
		//add static features
		
		// for each feature-sequence
		for (FeatureSequence fs : sequences) {
			int type = fs.TYPE;
			int[] windowLocations = this.getWindowLocationsForFeatureType(type);
			// for each sentence position
			for (int i:positions) { // iterate sentence positions
				for (int wloc: windowLocations) {
					int loc = i+wloc;					
					String symbol;
					if (loc < 0) { symbol = this.getPadding(loc, type); }
					else if (loc >= length) { symbol = this.getPadding(loc - length+1, type); }
					else { symbol = fs.symbols[loc]; }					
					featuresPerLocation[i].add(fmap.getFeatureId(symbol, type, wloc));
				}
			}				
		}
		// now each sentence position has all the static features 
				
		// decide each position, and then add the dynamic feature
		String[] dynamicSymbols = new String[length];
		int type = this.fmap.dynFeatureType();
		int[] dynWindowLocations = this.getWindowLocationsForFeatureType(type);
		String symbol;
		
		for (int i:positions) { // iterate sentence positions
			// get static features
			TIntHashSet features = featuresPerLocation[i];
			// add dynamic features
			for (int wloc : dynWindowLocations) {
				int loc = i+wloc;
				if (loc < 0) { symbol = this.getPadding(loc, type); }
				else if (loc > length) { symbol = this.getPadding(loc - length+1, type); }
				else { symbol = dynamicSymbols[loc]; } 
				features.add(fmap.getFeatureId(symbol, type, wloc));				
			}
			features.remove(-1);
			// make sorted features array 
			int[] fids = features.toArray();
			Arrays.sort(fids);
			String dec = model.classifyToLabels(fids);
			dynamicSymbols[i] = dec;
			res.add(dec);			
		}
		return res;
	}
	
	private String getPadding(int loc, int type) {
		// TODO: padder class		
		if (loc < 0) { return loc + "__BOS__"; }
		return "+" + loc + "__EOS__"; 
	}
	
	private int[] getWindowLocationsForFeatureType(int type) {
		// TODO: make a paramater? read from Yamcha file?
		return this.fmap.getPossibleWindowLocationsForFeatureType(type);		
	}

	////////////////////////////////////////////////////////////////
	interface PositionSeqMaker {
		public int[] range(int upper);
	}
	
	class Backward implements PositionSeqMaker {
		public int[] range(int upper) {
			int[] r = new int[upper+1];
			for (int i=0; i<r.length;i++) {r[i] = upper - i; }
			return r;
		}
	}
	class Forward implements PositionSeqMaker {
		public int[] range(int upper) {
			int[] r = new int[upper+1];
			for (int i=0; i<r.length;i++) {r[i] = i; }
			return r;
		}		
	}
	
	class FeatureSequence {
		
		private int TYPE;
		private String[] symbols;

		public FeatureSequence(int TYPE, String[] symbols) {
			this.TYPE = TYPE;
			this.symbols = symbols;
		}
	}

}
