# encoding: cp1255
#================================================================ 
#   Copyright 2008, Yoav Goldberg.
#   
#   This file is part of splitSVM.
# 
#    splitSVM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    splitSVM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with splitSVM.  If not, see http://www.gnu.org/licenses/.
#    
#================================================================
from svmlightlib import *
from interfaces import *
from util import _read_to_blank
from libsvmlib import LIBSVMClassifierData

import sys


def _read_header(lines_gen):
   header_lines = _read_to_blank(lines_gen)   
   res = {}
   for line in header_lines:
      (key, val) = line.split(":",1)
      res[key] = val
   return ("\n".join(header_lines), res)

TYPE = 0
WIN  = 1
LOC  = 2
WRD  = 3

class Feature:
   def __init__(self, idx,fstring):
      try:
         (ty,win,loc,wrd) = fstring.split(':',3)
      except:
         print "Err??:",fstring
         (ty,win,wrd) = fstring.split(':')
         loc = '-1'
      self._feat = (ty,win,loc,wrd)
      self._str = fstring
      self._idx = idx

   def feat(self): return self._feat
   def idx(self): return self._idx
   def is_lex(self): return self._feat[2] == '0'
   def __str__(self): return self._str
   def __repr__(self): return self.__str__()


class YamchaFeatureVector(SVMLightFeatureVector): 
   def __init__(self, vec_line):
      (ai, vec_info) = vec_line.split(None,1)
      ai = float(ai)
      vec_features = []
      for feat in vec_info.split():
         f = feat.split(":")[0]
         vec_features.append(f)
      feats = [int(f) for f in vec_features]
      SVMLightFeatureVector.__init__(self, ai, feats)
      self._vec_line = vec_line

   
def _read_features(lines_gen):
   feature_lines = _read_to_blank(lines_gen)
   feat_map = {}
   rev_feat_map = {}
   for line in feature_lines:
      try:
         (num,f) = line.split()
      except: continue
      feat_map[int(num)] = Feature(int(num),f)
      rev_feat_map[f] = int(num)
   return (feat_map, rev_feat_map)

class YamchaClassifierData(SVMLightClassifierData):
   def __init__(self, feat_map, head_lines, vecs, thrs = 0, ):
      SVMLightClassifierData.__init__(self, vecs, thrs)
      self._head = head_lines
      self._feat_map = feat_map

      self.set_true_label(self._head[0].split()[1])
      self.set_false_label(self._head[0].split()[2])

   def name(self): return self.true_label() + " vs. " + self.false_label()
   def is_pairwise(self): return self._head[0].split()[0] == 'TODO'  # TODO
   def is_ovr(self):      return self._head[0].split()[0] == 'MULTI_CLASS'

   def coef0(self): return 1.0
   def gamma(self): return 1.0
   def degree(self): return 2.0
   def kernel_type(self): return 'polynomial'

   @classmethod
   def fromLines(cls, feat_map, lines_gen): # TODO why am I passing feat_map around?
      lines = _read_to_blank(lines_gen)
      head_lines   = lines[:10]
      vector_lines = lines[10:]
      threshold = float(head_lines[-1].split("#")[0])
      vectors = [YamchaFeatureVector(line) for line in vector_lines]
      return cls(feat_map, head_lines, vectors, threshold)

class YamchaMulticlassModel(ClassifierDataInterace):  # TODO: very similar to LIBSVMClassfierData, refactor common parts
   def __init__(self, feat_map, is_ovr, SVs, map_cid_thr, labels_to_cids, cids_to_labels):
      fids = sorted(feat_map.keys())
      ClassifierDataInterace.__init__(self, SVs, None, fids, None, None)
      self._cid_to_thr = map_cid_thr
      self._labels_to_cids = labels_to_cids
      self._cids_to_labels = cids_to_labels

      self._fmap = feat_map
      self._is_ovr = is_ovr

   def is_ovr(self): return self._is_ovr
   def is_pairwise(self): return not self._is_ovr
   def features_map(self): return  self._fmap 

   def coef0(self): return 1.0
   def gamma(self): return 1.0
   def degree(self): return 2.0
   def kernel_type(self): return 'polynomial'


class YAMCHAAsMulticlassReader:
   """
   Read Yamcha model.  Returns a "multiclass" model, meaning all vectors in all
   SVM models are returned together, with a different ai value for each of the
   actual classifiers (this is economic if the same vector is shared between the
   classifiers, this way the kernel computation for this vector happens only
   once.)
   This is similar to the way we work with libsvm, but in Yamcha case it's
   slightly more annoying to create the classifier, because it's not built in in
   the model representation.
   """
   @classmethod
   def read(cls, filename):
      lines_gen = file(filename)

      orig_header, header = _read_header(lines_gen)
      feat_map,    rev_map= _read_features(lines_gen)

      classifiers = {}
      done = False
      classifier_ids = itertools.count(0)
      while not done:
         try:
            classifier = YamchaClassifierData.fromLines(feat_map, lines_gen)
            print "read classifer: %s" % classifier.name()
            classifiers[classifier_ids.next()] = classifier
         except StopIteration:
            print "Done reading classifiers"
            done = True

      ## now that we have the separate classifiers, build the combined one
      # We'll return them as a LIBSVMClassifierData (for no particular reason,
      # other than the code already exists)
      #
      # V1.0: NO sharing of SVs.  Multiple SVs will exist multiple times.
      #
      # what we need is:
      # @ classifier_id_to_thr: map: classifierID(int) --> threshold b(float)
      # @ veclist: a list of all SVs.  Each SV has features()->list(int) and ais()->list((cid, ai))
      # @ labels_to_cids: map : (label1,label2) --> classifierID(int)
      # @ cids_to_labels: reverse of the above
      # @ is_ovr : are we ovr (one-vs-rest) or pairwise
      classifier_id_to_thr = {}
      veclist = []
      labels_to_cids = {}
      cids_to_labels = {}
      all_labels = set()
      is_ovr = False
      for (cid, classifier) in classifiers.iteritems():
         is_ovr = classifier.is_ovr() # assuming they are all the same
         print "id:",cid, ") name:",classifier.name()
         print "\tthreshold:", classifier.thrs()
         classifier_id_to_thr[cid] = classifier.thrs()
         print "\ttrue :",classifier.true_label()
         print "\tfalse:",classifier.false_label()
         lbls = (classifier.true_label(), classifier.false_label())
         all_labels.update(lbls)
         labels_to_cids[lbls] = cid
         cids_to_labels[cid] = lbls
         vecs = classifier.vecs()
         print "\tnum SVs:",len(vecs)
         for vec in vecs:
            vec.addAI(vec.ai(), cid)
            veclist.append(vec)

      print "Total number of SVs:",len(veclist)
      # we have the data, now create the combined classifier
      # TODO: this is the same code as in  libsvmlib.py reader.  Refactor to
      # remove duplication
      classifier = YamchaMulticlassModel(feat_map, is_ovr, veclist, classifier_id_to_thr, labels_to_cids, cids_to_labels)
      classifier.set_kernel_type("polynomial") # TODO
      classifier.set_degree(2) # TODO
      classifier.set_gamma(1.0)
      classifier.set_coef0(1.0)

      classifier.nr_classes = 'TODO'
      classifier.nr_classifiers = len(classifier_id_to_thr.keys())
      classifier.labels = all_labels

      return classifier, feat_map





if __name__ == '__main__':
   YAMCHAAsMulticlassReader.read("nopunct.txtmodel")








